import 'dart:async';
import 'dart:mirrors';
import 'dart:isolate';
import 'dart:io';
import 'package:embla/application.dart';
import 'package:embla/http.dart';
import 'package:template_cache/cache.dart';
import 'package:template_cache/render.dart' as template_cache;
import 'package:path/path.dart' as path;

class ViewComposer {
  final Cache cache;
  final String templatesDirectory;

  ViewComposer(this.cache, this.templatesDirectory);

  Template render(String template, [Map<Symbol, dynamic> locals]) {
    return new Template(
        path.join(templatesDirectory, template),
        locals ?? {},
    );
  }
}

@proxy
class Template implements Responsible {
  final String _template;
  final Map<Symbol, dynamic> _locals;

  Template(this._template, this._locals);

  DataResponse toResponse() {
    final template = template_cache.view(_template, locals: _locals);
    final controller = new StreamController<String>();

    template.render().pipe(controller);

    return new DataResponse(
        controller.stream,
        ContentType.parse(template.contentType)
    );
  }

  noSuchMethod(Invocation invocation) {
    if (!invocation.isSetter) {
      return super.noSuchMethod(invocation);
    }

    final name = MirrorSystem
      .getName(invocation.memberName)
      .replaceFirst('=', '');

    _locals[new Symbol(name)] = invocation.positionalArguments[0];
  }
}

class TemplatingBootstrapper extends Bootstrapper {
  final Iterable compilers;
  final String templatesDirectory;
  Cache cache;

  TemplatingBootstrapper({
    this.compilers: const [],
    this.templatesDirectory: 'web'
  }) {
    cache = new Cache(_normalizeCompilers(compilers));
  }

  @Hook.bindings
  setUpViewComposer() {
    return container.bind(
        ViewComposer,
        to: new ViewComposer(cache, templatesDirectory)
    );
  }

  @Hook.init
  compile() async {
    final dir = new Directory(templatesDirectory);
    bool compiled = false;
    await for (final file in dir.list(recursive: true, followLinks: false)) {
      if (file is! File) continue;

      final filename = path.join(
          templatesDirectory,
          path.relative(file.path, from: dir.path)
      );

      if (await cache.shouldCompile(filename)) {
        await cache.compile(filename);
        print('<yellow>Compiled $filename</yellow>');
        compiled = true;
      }
    }
    if (compiled) {
      reload();
    }
  }

  Iterable<Compiler> _normalizeCompilers(Iterable compilers) sync* {
    for (final compiler in compilers) {
      if (compiler is Compiler) {
        yield compiler;
        continue;
      }

      if (compiler is Type) {
        final c = container.make(compiler);
        if (c is! Compiler) {
          throw new Exception('$compiler is not a template compiler.');
        }
        yield c;
      }
    }
  }
}
